/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#define FE_OTHR_DEBUG			FALSE
#define FE_OTHR_THREAD_VERBOSE	FALSE

#include "operate/operate.pmh"

using namespace fe;

OperatorThreaded::~OperatorThreaded(void)
{
	finish();
}

void OperatorThreaded::finish(void)
{
	for(U32 m=0;m<m_threadsAlive;m++)
	{
		m_spGang->post(sp<SurfaceRange>(NULL));
	}
	if(m_spGang->workers())
	{
		m_spGang->finish();
	}
	m_threadsAlive=0;
}

void OperatorThreaded::initialize(void)
{
	catalog<I32>("Threads")=1;
	catalog<I32>("Threads","min")=1;
	catalog<I32>("Threads","high")=16;
	catalog<bool>("Threads","joined")=true;
	catalog<String>("Threads","IO")="input output";
	catalog<String>("Threads","hint")=
			"Number of workers to process the data in parallel."
			"  The number of jobs may not be the same."
			"  When the thread count is set to one,"
			" no threads are spawned and the entire"
			" processing job is done at once with the existing thread.";

	catalog<bool>("AutoThread")=false;
	catalog<bool>("AutoThread","joined")=true;
	catalog<String>("AutoThread","IO")="input output";
	catalog<String>("AutoThread","hint")=
			"Automatically set the number of threads.";

	catalog<bool>("StayAlive")=false;
	catalog<String>("StayAlive","label")="Stay Alive";
	catalog<String>("StayAlive","hint")=
			"Keep the threads running and looking for jobs even"
			" when this node isn't being processed."
			"  This may speed up quick nodes where launching threads is"
			" a large portion of the processing time,"
			" but this technique should be considered tempermental"
			" (crashes may become more likely).";

	m_spGang=new Gang< Worker,sp<SurfaceRange> >();
}

void OperatorThreaded::adjustThreads(void)
{
	I32& rThreads=catalog<I32>("Threads");

	if(catalog<bool>("AutoThread"))
	{
		rThreads=boost::thread::hardware_concurrency();
	}

	setThreading((rThreads==1)? e_singleThread: e_multiThread);
}

void OperatorThreaded::handle(Record& a_rSignal)
{
#if FE_OTHR_DEBUG
	feLog("OperatorThreaded::handle \"%s\"\n",name().c_str());
#endif

	adjustThreads();

	I32& rThreads=catalog<I32>("Threads");
	const bool stayAlive=(catalog<bool>("StayAlive") && rThreads>1);

	if(m_threadsAlive &&
			(!stayAlive || (stayAlive && m_threadsAlive!=rThreads)))
	{
#if FE_OTHR_THREAD_VERBOSE
		feLog("Pre-finish stayAlive %d threadsAlive %d\n",
				stayAlive,m_threadsAlive);
#endif
		finish();
	}

	//* for a single thread, run directly
	if(rThreads==1)
	{
#if FE_OTHR_THREAD_VERBOSE
		feLog("OperatorThreaded::handle single thread [%d,%d]\n",
				m_spFullRange->brief().c_str());
#endif

		run(-1,m_spFullRange);
		return;
	}

	const U32 atomicCount=m_spFullRange->atomicCount();
	for(U32 atomicIndex=0;atomicIndex<atomicCount;atomicIndex++)
	{
		const SurfaceRange::MultiSpan& rAtomic=
				m_spFullRange->atomic(atomicIndex);
		sp<SurfaceRange> spSubRange(new SurfaceRange());
		spSubRange->addAtomic()=rAtomic;

#if FE_OTHR_THREAD_VERBOSE
		feLog("Post atomic %s\n",spSubRange->brief().c_str());
#endif
		m_spGang->post(spSubRange);
	}

	//* if atomics exist, presume non-atomic must be done serially
	if(!atomicCount)
	{
		const SurfaceRange::MultiSpan& rNonAtomic=m_spFullRange->nonAtomic();
		const U32 spanCount=rNonAtomic.count();

		for(U32 spanIndex=0;spanIndex<spanCount;spanIndex++)
		{
			const SurfaceRange::Span& rSpan=rNonAtomic.span(spanIndex);
			const U32 count=rSpan[1]-rSpan[0]+1;

			const U32 jobs=rThreads*4;	//* TODO param?
			const U32 each=(count>jobs)? (count+jobs-1)/jobs: 1;

			for(U32 m=0;m<count;m+=each)
			{
				const I32 start=rSpan[0]+m;
				I32 end=rSpan[0]+each-1;
				if(end>rSpan[1])
				{
					end>rSpan[1];
				}

#if FE_OTHR_THREAD_VERBOSE
				feLog("Post non-atomic [%d,%d]\n",start,end);
#endif
				if(end>=start)
				{
					sp<SurfaceRange> spSubRange(new SurfaceRange());
					spSubRange->nonAtomic().add(start,end);

					m_spGang->post(spSubRange);
				}
			}
		}
	}

	if(!stayAlive || !m_threadsAlive)
	{
#if FE_OTHR_THREAD_VERBOSE
		feLog("Start %d threads\n",rThreads);
#endif
		if(!m_spGang->start(rThreads,sp<OperatorThreaded>(this)))
		{
			feLog("OperatorThreaded::handle \"%s\" threading failed"
					" -> switching to single thread\n",name().c_str());

			m_threadsAlive=0;
			catalog<bool>("AutoThread")=FALSE;
			rThreads=1;
			setThreading(e_singleThread);
			run(-1,m_spFullRange);
			return;
		}
		if(stayAlive)
		{
			m_threadsAlive=rThreads;
		}
	}

	if(!stayAlive)
	{
#if FE_OTHR_THREAD_VERBOSE
		feLog("Post-finish stayAlive %d threadsAlive %d\n",
				stayAlive,m_threadsAlive);
#endif
		finish();
	}

	setThreading(e_singleThread);

	if(atomicCount)
	{
		const SurfaceRange::MultiSpan& rNonAtomic=
				m_spFullRange->nonAtomic();
		sp<SurfaceRange> spSubRange(new SurfaceRange());
		spSubRange->addAtomic()=rNonAtomic;

#if FE_OTHR_THREAD_VERBOSE
		feLog("Run serial %s\n",spSubRange->brief().c_str());
#endif
		run(-1,spSubRange);
	}

#if FE_OTHR_DEBUG
	feLog("OperatorThreaded::handle done\n");
#endif
}

void OperatorThreaded::Worker::operator()(void)
{
#if FE_OTHR_THREAD_VERBOSE
	feLog("Starting thread %d\n",m_id);
#endif
	const BWORD& wait=m_spOperatorThreaded->m_threadsAlive;
	BWORD done=FALSE;
	while(!done)
	{
		sp<SurfaceRange> spRange;

		if(wait)
		{
#if FE_OTHR_THREAD_VERBOSE
			feLog("Thread %d waiting\n",m_id);
#endif
			m_spJobQueue->waitForJob(spRange);
		}
		else
		{
			done=!m_spJobQueue->take(spRange);
		}

#if FE_OTHR_THREAD_VERBOSE
		feLog("Thread %d Job %s\n",m_id,
				spRange.isValid()? spRange->brief().c_str(): "NULL");
#endif

		if(spRange.isNull() || spRange->empty())
		{
			done=TRUE;
		}
		else if(!done)
		{
			m_spOperatorThreaded->run(m_id,spRange);
		}
	}
#if FE_OTHR_THREAD_VERBOSE
	feLog("Sync thread %d\n",m_id);
#endif
	U32 leader=m_spJobQueue->synchronize();
#if FE_OTHR_THREAD_VERBOSE
	feLog("Sync'd thread %d leader=%d\n",m_id,leader);
#endif
}
