/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#define FE_CVO_DEBUG		FALSE
#define FE_CVO_EDIT_DEBUG	FALSE

#include "operator/operator.pmh"

using namespace fe;

void CurvaceousOp::initialize(void)
{
	catalog<String>("icon")="FE_alpha";

	catalog<Real>("pickRadius")=0.0;
	catalog<Real>("pickRadius","high")=1.0;
	catalog<String>("pickRadius","label")="Pick Radius";
	catalog<String>("pickRadius","page")="Edit";
	catalog<String>("pickRadius","hint")="Inner region of area select.";

	catalog<Real>("pickFalloff")=1.0;
	catalog<String>("pickFalloff","label")="Pick Falloff";
	catalog<String>("pickFalloff","page")="Edit";
	catalog<String>("pickFalloff","hint")="Falloff extension of area select.";

	catalog<Real>("uRange")=0.0;
	catalog<Real>("uRange","high")=1.0;
	catalog<String>("uRange","label")="Unit CV Range";
	catalog<String>("uRange","page")="Edit";
	catalog<String>("uRange","hint")="Inner region of cv select.";

	catalog<Real>("uFalloff")=1.0;
	catalog<String>("uFalloff","label")="Unit CV Falloff";
	catalog<String>("uFalloff","page")="Edit";
	catalog<String>("uFalloff","hint")="Falloff extension of cv select.";

	catalog< sp<Component> >("Brush");
	catalog<String>("Brush","implementation")="DrawI";
	catalog<String>("Brush","visible")="any";
	catalog<String>("Brush","prompt")="Prompt here.";

	catalog< sp<Component> >("Input Surface");

	catalog< sp<Component> >("Driver Surface");
	catalog<bool>("Driver Surface","optional")=true;

	catalog< sp<Component> >("Output Surface");
	catalog<String>("Output Surface","IO")="output";
	catalog<String>("Output Surface","copy")="Input Surface";
	catalog<bool>("Output Surface","recycle")=true;

	m_spSolid=new DrawMode();
	m_spSolid->setDrawStyle(DrawMode::e_solid);
	m_spSolid->setLit(FALSE);
	m_spSolid->setPointSize(3.0);
	m_spSolid->setLineWidth(3.0);

	//* nested operator
	m_spLengthCorrect=registry()->create("OperatorSurfaceI.LengthCorrectOp");
	sp<Catalog> spOtherCatalog=m_spLengthCorrect;
	if(spOtherCatalog.isValid())
	{
		spOtherCatalog->catalog<bool>("Root Lock")=true;
		spOtherCatalog->catalog<I32>("Iterations")=20;
	}
}

void CurvaceousOp::bind(sp<SignalerI> a_spSignalerI,sp<Layout> a_spLayout)
{
	OperatorSurfaceCommon::bind(a_spSignalerI,a_spLayout);

	if(m_spLengthCorrect.isValid())
	{
		m_spLengthCorrect->bind(a_spSignalerI,a_spLayout);
	}
}

void CurvaceousOp::handle(Record& a_rSignal)
{
	const Real pickRadius=catalog<Real>("pickRadius");
	const Real pickFalloff=catalog<Real>("pickFalloff");
	const Real uRange=catalog<Real>("uRange");
	const Real uFalloff=catalog<Real>("uFalloff");

	sp<DrawI> spDrawBrush;
	accessBrush(spDrawBrush,a_rSignal,e_quiet);

	sp<DrawI> spDrawOverlay;
	accessBrushOverlay(spDrawOverlay,a_rSignal,e_quiet);

#if FE_CVO_DEBUG
	feLog("CurvaceousOp::handle brush %d brushed %d frame %.6G\n",
			spDrawBrush.isValid(),m_brushed,m_frame);
#endif

	if(spDrawBrush.isValid())
	{
#if FALSE
		I32 pickPrimitive=lookupIndex(m_pickId);
		if(pickPrimitive>=m_spOutputVertices->count())
		{
			pickPrimitive= -1;
		}
		if(pickPrimitive<0)
		{
			m_pickId= -1;
			m_pickVertex= -1;
			m_pickU=0.0;
		}
		I32 pickVertex=m_pickVertex;
		Real pickU=m_pickU;
#endif

		//* HACK
		I32 pickPrimitive= -1;
		I32 pickVertex= -1;
		Real pickU=0.0;

		spDrawBrush->pushDrawMode(m_spSolid);

		m_brushed=TRUE;

		const Color black(0.0,0.0,0.0);
		const Color white(1.0,1.0,0.8);
		const Color grey(0.8,0.8,1.0);
		const Color darkgrey(0.5,0.5,0.5);
		const Color red(0.8,0.0,0.0);
		const Color green(0.0,0.5,0.0);
		const Color blue(0.0,0.0,1.0);
		const Color cyan(0.0,1.0,1.0);
		const Color magenta(1.0,0.0,1.0);
		const Color pink(0.8,0.4,0.4);
		const Color yellow(1.0,1.0,0.0);

		const SpatialVector up(0.0,1.0,0.0);

		sp<ViewI> spView=spDrawOverlay->view();
		sp<ViewMode> spViewMode=spView->viewMode();
		const Box2i viewport=spViewMode->viewport();
		SpatialTransform camera=spViewMode->camera();
		SpatialTransform invCamera;
		invert(invCamera,camera);

		const SpatialVector deltaX=invCamera.column(0);
		const SpatialVector deltaY=invCamera.column(1);

		if(m_spOutput.isValid())
		{
			const SpatialVector& rRayOrigin=rayOrigin(a_rSignal);
			const SpatialVector& rRayDirection=rayDirection(a_rSignal);

			m_event.bind(windowEvent(a_rSignal));

#if FE_CVO_DEBUG
			feLog("%s\n",c_print(m_event));
			feLog("ray %s  %s\n",c_print(rRayOrigin),c_print(rRayDirection));
#endif

			const Real maxDistance=100.0;
			sp<SurfaceI::ImpactI> spImpact=m_spOutput->rayImpact(
					rRayOrigin,rRayDirection,maxDistance);

			if(spImpact.isValid())
			{
				sp<SurfaceSearchable::Impact> spSearchableImpact=spImpact;
				if(spSearchableImpact.isValid())
				{
					pickPrimitive=spSearchableImpact->pickIndex();
					pickU=spSearchableImpact->uv()[0];

					const I32 subCount=
							m_spOutputVertices->subCount(pickPrimitive);
					pickVertex=pickU*(subCount-1.0)+0.499;
				}
			}

			U32 item=m_event.item();
			BWORD shifted=FALSE;
			if(item&WindowEvent::e_keyShift)
			{
				shifted=TRUE;
				item^=WindowEvent::e_keyShift;
			}
			BWORD ctrled=FALSE;
			if(item&WindowEvent::e_keyControl)
			{
				ctrled=TRUE;
				item^=WindowEvent::e_keyControl;
			}

			SpatialTransform pickedTip;

			if(m_event.isKeyPress(WindowEvent::e_itemAny))
			{
				if(item==WindowEvent::e_keyEscape)
				{
					m_pickId= -1;
					m_pickVertex= -1;
					m_pickU=0.0;
					pickPrimitive= -1;
				}
				if(item==WindowEvent::e_keyCarriageReturn)
				{
				}
				if(item==WindowEvent::e_keyDelete)
				{
				}
			}

			if(!m_dragging && m_event.isMouseRelease(WindowEvent::e_itemLeft))
			{
				m_pickId=lookupId(pickPrimitive);
				m_pickVertex=pickVertex;
				m_pickU=pickU;
				if(m_pickId<0)
				{
					m_pickVertex= -1;
					m_pickU=0.0;
				}
			}

			const WindowEvent::MouseButtons buttons=m_event.mouseButtons();
			if(buttons)
			{
				//* TODO screen aware
				const Real moveScale=1e-2;

				const I32 mouseX=m_event.mouseX();
				const I32 mouseY=m_event.mouseY();

				const I32 moveX=mouseX-m_lastX;
				const I32 moveY=mouseY-m_lastY;

				m_lastX=mouseX;
				m_lastY=mouseY;
				m_dragging=m_event.isMouseDrag();

				if(m_dragging)
				{
					const SpatialVector delta=
							moveScale*(deltaX*moveX+deltaY*moveY);

					movePicked(delta);
				}
			}
		}

		if(spDrawOverlay.isValid())
		{
			spDrawOverlay->pushDrawMode(m_spSolid);

			if(m_pickId>=0)
			{
				pickPrimitive=lookupIndex(m_pickId);
				pickVertex=m_pickVertex;
				pickU=m_pickU;

				m_pickArray.clear();	//* TODO avoid redundant repeats
			}

			if(pickPrimitive>=0)
			{
				highlightCurve(spView,spDrawOverlay,m_spOutputVertices,
						pickPrimitive,pickVertex,&cyan,&black,&yellow);

				const SpatialVector searchPoint=
						m_spOutputVertices->spatialVector(pickPrimitive,0);
				const SpatialVector pickPoint=
						m_spOutputVertices->spatialVector(
						pickPrimitive,pickVertex);

				const Real extent=pickRadius+pickFalloff;
				const Real extent2=extent*extent;

				const Real range=uRange+uFalloff;
				const Real unitVertex=pickVertex/
							(m_spOutputVertices->subCount(pickPrimitive)-1.0);

				if(m_spDriver.isValid())
				{
					sp<SurfaceI::ImpactI> spImpact=
							m_spDriver->nearestPoint(searchPoint);
					if(spImpact.isValid())
					{
						const Vector2 uv=spImpact->uv();
						SpatialTransform xform=m_spDriver->sample(uv);
						rotate(xform,-90.0*degToRad,e_xAxis);

						const SpatialVector circleScale(extent,
								extent,extent);

						spDrawBrush->drawCircle(xform,&circleScale,green);
					}
				}

				const U32 primitiveCount=m_spOutputVertices->count();
				for(U32 primitiveIndex=0;primitiveIndex<primitiveCount;
						primitiveIndex++)
				{
					const SpatialVector rootPoint=
							m_spOutputVertices->spatialVector(primitiveIndex,0);
					const Real dist2=magnitudeSquared(rootPoint-searchPoint);

					if(dist2>extent2)
					{
						continue;
					}

					const Real dist=sqrt(dist2);
					Real primitiveWeight=1.0;
					if(primitiveIndex!=pickPrimitive &&
							pickFalloff>0.0 && dist>pickRadius)
					{
						primitiveWeight-=(dist-pickRadius)/pickFalloff;
					}
					const Color blueish(0.0,primitiveWeight,
							0.3+0.7*primitiveWeight);

					highlightCurve(spView,spDrawOverlay,m_spOutputVertices,
							primitiveIndex,-1,&blueish,&black,NULL);

					const I32 subCount=
							m_spOutputVertices->subCount(primitiveIndex);
					for(I32 subIndex=1;subIndex<subCount;subIndex++)
					{
						const SpatialVector point=
								m_spOutputVertices->spatialVector(
								primitiveIndex,subIndex);

						const Real unitIndex=subIndex/(subCount-1.0);
						const Real pointDist=fabs(unitIndex-unitVertex);

						if(pointDist>range)
						{
							continue;
						}

						Real pointWeight=1.0;
						if(!(primitiveIndex==pickPrimitive &&
								subIndex==pickVertex) &&
								uFalloff>0.0 && pointDist>uRange)
						{
							pointWeight-=(pointDist-uRange)/uFalloff;
						}

						const Real productWeight=primitiveWeight*pointWeight;
						if(productWeight<0.01)	//* tweak
						{
							continue;
						}

						const Color reddish(1.0,productWeight,0.0);

						drawDot(spView,spDrawOverlay,point,3.0,reddish);

						const I32 pointIndex=
								m_spOutputVertices->integer(
								primitiveIndex,subIndex);
						pick(lookupId(pointIndex),productWeight);
					}
				}
			}

			spDrawOverlay->popDrawMode();
		}

		spDrawBrush->popDrawMode();

		return;
	}

	access(m_spInputPoint,"Input Surface",e_point,e_position,e_quiet);

	access(m_spInputVertices,"Input Surface",e_primitive,e_vertices,e_quiet);

	if(!accessOutput(m_spOutputVertices,a_rSignal,
			e_primitive,e_vertices)) return;

	if(!accessOutput(m_spOutputPoint,a_rSignal,
			e_point,e_position)) return;

	if(!accessOutput(m_spOutput,a_rSignal)) return;

	access(m_spDriver,"Driver Surface",e_quiet);

	m_frame=currentFrame(a_rSignal);

	const BWORD frameChanged=(m_frame!= m_lastFrame);
	const BWORD replaced=(catalog<bool>("Output Surface","replaced"));
	const BWORD paramChanged=(!m_brushed && !replaced && !frameChanged);
	m_brushed=FALSE;

	const BWORD rebuild=(paramChanged || frameChanged || replaced);

#if FE_CVO_DEBUG
	feLog("CurvaceousOp::handle frame %.6G->%.6G paramChanged %d"
			" frameChanged %d replaced %d\n",
			m_lastFrame,m_frame,paramChanged,frameChanged,replaced);
#endif

	m_lastFrame=m_frame;

	if(m_spLengthCorrect.isValid())
	{
		sp<Catalog> spOtherCatalog=m_spLengthCorrect;
		if(spOtherCatalog.isValid())
		{
			spOtherCatalog->catalog< sp<Component> >("Input Surface")=
					catalog< sp<Component> >("Input Surface");
		}
		m_spLengthCorrect->handle(a_rSignal);
	}
}

void CurvaceousOp::setupState(void)
{
	m_aIndex.populate(m_spScope,"Move","Index");
	m_aOffset.populate(m_spScope,"Move","Offset");

	m_spEditMap=m_spState->lookupMap(m_aIndex);
}

BWORD CurvaceousOp::loadState(const String& rBuffer)
{
	const BWORD success=OperatorState::loadState(rBuffer);

	m_spEditMap=m_spState->lookupMap(m_aIndex);

	return success;
}

void CurvaceousOp::movePicked(const SpatialVector& a_rDelta)
{
#if FE_CVO_EDIT_DEBUG
	feLog("CurvaceousOp::movePicked %s\n",c_print(a_rDelta));
#endif

	const I32 pickCount=m_pickArray.size();
	for(I32 pickIndex=0;pickIndex<pickCount;pickIndex++)
	{
		const Pick& rPick=m_pickArray[pickIndex];
		const I32 pointIndex=rPick.m_index;
		const Real pointWeight=rPick.m_weight;

		const U32 count=m_spOutputPoint->count();
		if(pointIndex>=count)
		{
			continue;
		}

		SpatialVector point=m_spOutputPoint->spatialVector(pointIndex);
		point+=pointWeight*a_rDelta;

		m_spOutputPoint->set(pointIndex,point);

#if FE_CVO_EDIT_DEBUG
		feLog("CurvaceousOp::movePicked %d %d at %s to %s\n",
				pickIndex,pointIndex,c_print(point));
#endif
	}
}
