/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

//* NOTE refinement with multithreading will crash

/** @file */

#define FE_LCO_DEBUG	FALSE

#include "operator/operator.pmh"

using namespace fe;

void LengthCorrectOp::initialize(void)
{
	catalog<I32>("Iterations")=1;
	catalog<I32>("Iterations","high")=10;
	catalog<I32>("Iterations","max")=100;

	catalog<bool>("Root Lock")=false;
	catalog<String>("Root Lock","hint")=
			"Leave first vertex of each primitive untouched."
			"  This can lock down rooted curves, like hair.";

	catalog<Real>("Tension Rate")=1.0;
	catalog<Real>("Tension Rate","high")=1.0;
	catalog<String>("Tension Rate","hint")=
			"Influence how quickly each iteration pulls towards"
			" the correct length.";

	catalog<Real>("Compression Rate")=1.0;
	catalog<Real>("Compression Rate","high")=1.0;
	catalog<String>("Compression Rate","hint")=
			"Influence how quickly each iteration pushes towards"
			" the correct length.";

	catalog<Real>("Correction Bias")=0.5;
	catalog<Real>("Correction Bias","max")=1.0;
	catalog<String>("Correction Bias","hint")=
			"Influence how much each point on an edge gets changed."
			"  Try lower numbers for rooted curves, like 0.3,"
			" but generally leave at 0.5 for polygons.";

	catalog<Real>("Suspension")=0.0;
	catalog<Real>("Suspension","max")=1.0;
	catalog<String>("Suspension","hint")=
			"Restrain changes to slide along original input.";

	catalog<bool>("Collide")=false;
	catalog<String>("Collide","label")="Collision Avoidance";
	catalog<bool>("Collide","joined")=true;
	catalog<String>("Collide","hint")=
			"Push resulting surface out of collider.";

	catalog<Real>("Margin")=0.1;
	catalog<bool>("Margin","joined")=true;
	catalog<Real>("Margin","high")=1.0;
	catalog<Real>("Margin","max")=1e3;
	catalog<String>("Margin","enabler")="Collide";
	catalog<String>("Margin","hint")=
			"Minimum distance to keep from collider surface.";

	catalog<Real>("CollideRate")=1.0;
	catalog<String>("CollideRate","label")="Rate";
	catalog<Real>("CollideRate","high")=1.0;
	catalog<String>("CollideRate","enabler")="Collide";
	catalog<String>("CollideRate","hint")=
			"Amount of collision response per iteration.";

	catalog<I32>("CollideRefinement")=0;
	catalog<String>("CollideRefinement","label")="Refinement";
	catalog<String>("CollideRefinement","enabler")="Collide";

	catalog<bool>("Locking")=false;
	catalog<bool>("Locking","joined")=true;
	catalog<String>("Locking","hint")=
			"Use locking attribute to resist change.";

	catalog<String>("LockAttr")="lock";
	catalog<String>("LockAttr","label")="Lock Attribute";
	catalog<String>("LockAttr","enabler")="Locking";
	catalog<String>("LockAttr","hint")=
			"Name of real attribute describing resistance to change"
			" (1.0 is full lock).";

	catalog< sp<Component> >("Input Surface");

	catalog< sp<Component> >("Reference Surface");

	catalog< sp<Component> >("Collider Surface");
	catalog<bool>("Collider Surface","optional")=true;
}

void LengthCorrectOp::handle(Record& a_rSignal)
{
#if FE_LCO_DEBUG
	feLog("LengthCorrectOp::handle\n");
#endif

	const I32 iterations=catalog<I32>("Iterations");
	const BWORD rootLock=catalog<bool>("Root Lock");
	const Real tensionRate=catalog<Real>("Tension Rate");
	const Real compressionRate=catalog<Real>("Compression Rate");
	const Real correctionBias=catalog<Real>("Correction Bias");
	const Real suspension=catalog<Real>("Suspension");
	const String lockAttr=catalog<String>("LockAttr");
	const BWORD locking=(catalog<bool>("Locking") && !lockAttr.empty());
	const BWORD collide=catalog<bool>("Collide");

	//* output surface is copy of input surface
	if(!accessOutput(m_spOutputAccessible,a_rSignal)) return;

	//* access reference surface
	sp<SurfaceAccessibleI> spReferenceAccessible;
	if(!access(spReferenceAccessible,"Reference Surface",e_quiet))
	{
		if(!access(spReferenceAccessible,"Input Surface")) return;

		catalog<String>("warning")=
				"Substituting Input Surface for missing Reference Surface.";
	}

	sp<SurfaceAccessorI> spInputLock;
	if(locking)
	{
		access(spInputLock,"Input Surface",e_point,lockAttr,e_warning);
	}

	sp<SurfaceAccessorI> spInputOpen;
	if(!access(spInputOpen,"Input Surface",e_primitive,e_open)) return;

	sp<SurfaceAccessorI> spOutputVertices;
	if(!access(spOutputVertices,m_spOutputAccessible,
			e_primitive,e_vertices)) return;

	sp<SurfaceAccessorI> spRefVertices;
	if(!access(spRefVertices,spReferenceAccessible,
			e_primitive,e_vertices)) return;

	//* TODO need special case for curves
	if(suspension>0.0)
	{
		//* access input surface
		sp<SurfaceAccessibleI> spInputAccessible;
		if(!access(spInputAccessible,"Input Surface")) return;

		if(!access(m_spDriver,spInputAccessible)) return;

		m_spDriver->setRefinement(0);
		m_spDriver->prepareForSearch();
	}
	else
	{
		m_spDriver=NULL;
	}

	m_spCollider=NULL;
	if(collide)
	{
		access(m_spCollider,"Collider Surface",e_warning);
		if(m_spCollider.isValid())
		{
			m_spCollider->setRefinement(catalog<I32>("CollideRefinement"));
			m_spCollider->prepareForSearch();
		}
	}

	//* tweak surface points towards reference length
	const U32 count=spOutputVertices->count();
	const U32 countRef=spRefVertices->count();
	if(count!=countRef)
	{
		feLog("LengthCorrectOp::handle mismatched count %d vs %d ref\n",
				count,countRef);
		catalog<String>("error")="mismatched reference surface";
		return;
	}

#if FE_LCO_DEBUG
	feLog("LengthCorrectOp::handle primitives %d\n", count);
#endif

	for(U32 iteration=0;iteration<iterations;iteration++)
	{
		for(U32 index=0;index<count;index++)
		{
			const BWORD openCurve=spInputOpen->integer(index);
			const U32 subCount=spOutputVertices->subCount(index);
			const U32 subCountRef=spRefVertices->subCount(index);
			if(subCount!=subCountRef)
			{
				feLog("LengthCorrectOp::handle"
						" primitive %d mismatched count %d vs %d ref\n",
						index,subCount,subCountRef);
				continue;
			}

#if FE_LCO_DEBUG
			feLog("LengthCorrectOp::handle iteration %d/%d"
					" primitive %d/%d subCount %d\n",
					iteration,iterations,index,count, subCount);
#endif

			//* TODO check for wrap-around (closed poly vs open curve)

			//* NOTE skip root point with rootLock or any open curves
			for(U32 subIndex=(rootLock||openCurve);
					subIndex<subCount;subIndex++)
			{
				const U32 lastSubIndex=subIndex? subIndex-1: subCount-1;
				SpatialVector pointA=
						spOutputVertices->spatialVector(index,lastSubIndex);
				SpatialVector pointB=
						spOutputVertices->spatialVector(index,subIndex);

#if FE_LCO_DEBUG
				feLog("LengthCorrectOp::handle   %d/%d pointA %s pointB %s\n",
						subIndex,subCount,c_print(pointA),c_print(pointB));
#endif

				const SpatialVector delta=pointB-pointA;
				const Real length=magnitude(delta);
				if(length>0.0)
				{
					Real lockA=0.0;
					Real lockB=0.0;
					if(spInputLock.isValid())
					{
						const I32 vertexIndexA=
								spOutputVertices->integer(index,lastSubIndex);
						const I32 vertexIndexB=
								spOutputVertices->integer(index,subIndex);
						lockA=spInputLock->real(vertexIndexA);
						lockB=spInputLock->real(vertexIndexB);
					}
					const BWORD rootLocked=(rootLock && subIndex<2);
					if(rootLocked)
					{
						lockA=1.0;
					}

					const Real lockBias=0.5*(1.0+lockB-lockA);
					Real bias=lockBias+correctionBias-0.5;
					if(bias<0.0)
					{
						bias=0.0;
					}
					if(bias>1.0)
					{
						bias=1.0;
					}
					const Real antibias=1.0-bias;

					const SpatialVector pointRefA=
							spRefVertices->spatialVector(index,lastSubIndex);
					const SpatialVector pointRefB=
							spRefVertices->spatialVector(index,subIndex);
					const Real lengthRef=magnitude(pointRefB-pointRefA);

					const Real rate=(length>lengthRef)?
							tensionRate: compressionRate;
					const SpatialVector adjust=
							(lengthRef/length-1.0)*delta*rate;

#if FE_LCO_DEBUG
					feLog("LengthCorrectOp::handle   ref pointA %s pointB %s\n",
							c_print(pointRefA),c_print(pointRefB));
#endif

					if(rootLocked)
					{
						pointB+=(1.0-lockB)*adjust;
						spOutputVertices->set(index,subIndex,pointB);
					}
					else
					{
						pointA-=(1.0-lockA)*bias*adjust;
						pointB+=(1.0-lockB)*antibias*adjust;

						spOutputVertices->set(index,lastSubIndex,pointA);
						spOutputVertices->set(index,subIndex,pointB);
					}

#if FE_LCO_DEBUG
					feLog("LengthCorrectOp::handle"
							"   bias %.6G lock pointA %.6G pointB %.6G\n",
							bias,lockA,lockB);
					feLog("LengthCorrectOp::handle   now pointA %s pointB %s\n",
							c_print(pointA),c_print(pointB));
#endif
				}
			}
		}

		if(m_spDriver.isValid() || m_spCollider.isValid())
		{
			setAtomicRange(m_spOutputAccessible,e_pointsOnly);
			OperatorThreaded::handle(a_rSignal);

		}
	}

#if FE_LCO_DEBUG
	feLog("LengthCorrectOp::handle done\n");
#endif
}

void LengthCorrectOp::run(I32 a_id,sp<SurfaceRange> a_spRange)
{
	const Real suspension=catalog<Real>("Suspension");
	const String lockAttr=catalog<String>("LockAttr");
	const BWORD locking=(catalog<bool>("Locking") && !lockAttr.empty());
	const Real margin=catalog<Real>("Margin");
	const Real collideRate=catalog<Real>("CollideRate");

	sp<SurfaceAccessorI> spOutputPoints;
	if(!access(spOutputPoints,m_spOutputAccessible,
			e_point,e_position)) return;

	sp<SurfaceAccessorI> spInputLock;
	if(locking)
	{
		access(spInputLock,"Input Surface",e_point,lockAttr,e_quiet);
	}

	for(SurfaceRange::Iterator it=a_spRange->begin();!it.atEnd();it.step())
	{
		const U32 index=it.value();
		SpatialVector point=spOutputPoints->spatialVector(index);
		Real lock=0.0;
		if(spInputLock.isValid())
		{
			lock=spInputLock->real(index);
			if(lock>=1.0)
			{
				continue;
			}
		}

		BWORD changed=FALSE;

		if(m_spDriver.isValid())
		{
			sp<SurfaceI::ImpactI> spImpact=m_spDriver->nearestPoint(point);
			if(spImpact.isValid())
			{
				point+=(1.0-lock)*suspension*(spImpact->intersection()-point);
				changed=TRUE;
			}
		}

		if(m_spCollider.isValid())
		{
			std::vector< sp<SurfaceI::ImpactI> > impactArray=
					m_spCollider->nearestPoints(point,1);
			for(U32 m=0;m<impactArray.size();m++)
			{
				sp<SurfaceI::ImpactI> spImpact=impactArray[m];
				const SpatialVector norm=spImpact->normal();

				const SpatialVector target=
						spImpact->intersection()+margin*norm;

				const SpatialVector toTarget=target-point;
				const Real along=dot(toTarget,norm);
				if(along>0.0)
				{
#if FALSE
					feLog("LengthCorrectOp::run"
							"   point %s int %s\n norm %s along %.6G\n",
							c_print(point),c_print(spImpact->intersection()),
							c_print(norm),along);
#endif

					point+=collideRate*(1.0-lock)*along*norm;
					changed=TRUE;

#if FALSE
					feLog("   target %s to %s\n",
							c_print(toTarget),c_print(point));
#endif
				}
			}
		}

		if(changed)
		{
			spOutputPoints->set(index,point);
		}
	}
}
